package character

import (
	"encoding/json"
	"errors"

	functionset "example.com/database/functionset"
	inventoryslot "example.com/database/inventoryslot"
	person "example.com/database/person"
	role "example.com/database/role"

	"gorm.io/gorm"
)

type Character struct {
	gorm.Model
	Name         string                        `gorm:"uniqueIndex" json:"name"`
	Owners       []person.Person               `gorm:"many2many:character_owner_associations" json:"owners"` // Unique
	Roles        []role.Role                   `gorm:"many2many:character_role_associations" json:"roles"`   // Unique
	FunctionSets []functionset.FunctionSet     `gorm:"many2many:character_functionset_associations" json:"function_sets"`
	Inventory    []inventoryslot.InventorySlot `gorm:"many2many:character_inventory_associations" json:"inventory"`
}

type CharacterParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	Name         string `json:"name"`
	Owners       []uint `json:"owners"`
	Roles        []uint `json:"roles"`
	FunctionSets []uint `json:"function_sets"`
	Inventory    []uint `json:"inventory_slot"`
}

func (params *CharacterParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams CharacterParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	if inputParams.Name == "" {
		return errors.New("input name cannot be empty")
	}
	params.Name = inputParams.Name
	params.Owners = inputParams.Owners
	params.Roles = inputParams.Roles
	params.FunctionSets = inputParams.FunctionSets
	params.Inventory = inputParams.Inventory
	return nil
}

func (character Character) create(db *gorm.DB) error {
	result := db.Create(&character)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (character *Character) getAssociations(db *gorm.DB) {
	db.Model(&character).Association("Owners").Find(&character.Owners)
	db.Model(&character).Association("Roles").Find(&character.Roles)
	db.Model(&character).Association("FunctionSets").Find(&character.FunctionSets)
	db.Model(&character).Association("Inventory").Find(&character.Inventory)
}

func (character *Character) get(db *gorm.DB, inputCharacter uint) {
	db.Where("id = ?", inputCharacter).Take(&character)
	character.getAssociations(db)
}

func (character Character) update(db *gorm.DB) error {
	// Get the original Character object
	var originalCharacter Character
	result := db.Model(&Character{}).Where("id = ?", character.Model.ID).Take(&originalCharacter)
	if result.Error != nil {
		return result.Error
	}
	// Set the static values
	originalCharacter.Name = character.Name
	// Set the associated values by grabbing them from the database
	ownersError := db.Model(&originalCharacter).Association("Owners").Replace(&character.Owners)
	if ownersError != nil {
		return ownersError
	}
	rolesError := db.Model(&originalCharacter).Association("Roles").Replace(&character.Roles)
	if rolesError != nil {
		return rolesError
	}
	functionSetsError := db.Model(&originalCharacter).Association("FunctionSets").Replace(&character.FunctionSets)
	if functionSetsError != nil {
		return functionSetsError
	}
	inventoryError := db.Model(&originalCharacter).Association("Inventory").Replace(&character.Inventory)
	if inventoryError != nil {
		return inventoryError
	}
	result = db.Model(&originalCharacter).Update("name", originalCharacter.Name)
	if result.Error != nil {
		err := errors.New("new name already exists")
		return err
	}
	db.Save(&originalCharacter)
	return nil
}

func (character Character) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&character)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params CharacterParams) error {
	var newOwners []person.Person
	if len(params.Owners) > 0 {
		newOwners = *person.Get(db, params.Owners)
	}
	var newRoles []role.Role
	if len(params.Roles) > 0 {
		newOwners = *person.Get(db, params.Roles)
	}
	var newFunctionSets []functionset.FunctionSet
	if len(params.FunctionSets) > 0 {
		newOwners = *person.Get(db, params.FunctionSets)
	}
	var newInventory []inventoryslot.InventorySlot
	if len(params.Inventory) > 0 {
		newOwners = *person.Get(db, params.Inventory)
	}
	return Character{
		Name:         params.Name,
		Owners:       newOwners,
		Roles:        newRoles,
		FunctionSets: newFunctionSets,
		Inventory:    newInventory,
	}.create(db)
}

func Get(db *gorm.DB, inputCharacters []uint) *[]Character {
	var outputCharacters []Character
	if len(inputCharacters) < 1 {
		db.Model(&Character{}).Select("id").Find(&inputCharacters)
	}
	for _, inputCharacter := range inputCharacters {
		var outputCharacter Character
		outputCharacter.get(db, inputCharacter)
		outputCharacters = append(outputCharacters, outputCharacter)
	}
	return &outputCharacters
}

func Update(db *gorm.DB, params CharacterParams) error {
	var newOwners []person.Person
	if len(params.Owners) > 0 {
		newOwners = *person.Get(db, params.Owners)
	}
	var newRoles []role.Role
	if len(params.Roles) > 0 {
		newOwners = *person.Get(db, params.Roles)
	}
	var newFunctionSets []functionset.FunctionSet
	if len(params.FunctionSets) > 0 {
		newOwners = *person.Get(db, params.FunctionSets)
	}
	var newInventory []inventoryslot.InventorySlot
	if len(params.Inventory) > 0 {
		newOwners = *person.Get(db, params.Inventory)
	}
	return Character{
		Name:         params.Name,
		Owners:       newOwners,
		Roles:        newRoles,
		FunctionSets: newFunctionSets,
		Inventory:    newInventory,
	}.update(db)
}

func Delete(db *gorm.DB, inputCharacters []uint) error {
	var characters []Character
	// if len(inputCharacters) < 1 {
	// 	result := db.Model(&Character{}).Select("id").Find(&inputCharacters)
	// 	if result.Error != nil {
	// 		return result.Error
	// 	}
	// }
	for _, inputCharacter := range inputCharacters {
		var character Character
		character.get(db, inputCharacter)
		characters = append(characters, character)
	}
	for _, character := range characters {
		err := character.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) (*[]Character, error) {
	var err error
	var params CharacterParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result *[]Character
	switch method {
	case "GET":
		result = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
