package function

import (
	"encoding/json"
	"errors"

	functiontag "example.com/database/functiontag"

	"gorm.io/gorm"
)

type Function struct {
	gorm.Model
	Name         string                    `gorm:"uniqueIndex" json:"name"`
	Tags         []functiontag.FunctionTag `gorm:"many2many:function_tag_associations" json:"tags"`
	Requirements []Function                `gorm:"many2many:function_requirement_associations" json:"requirements"`
}

type FunctionParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	Name         string `json:"name"`
	Tags         []uint `json:"tags"`
	Requirements []uint `json:"requirements"`
}

func (params *FunctionParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams FunctionParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	if inputParams.Name == "" {
		return errors.New("input name cannot be empty")
	}
	params.Name = inputParams.Name
	params.Tags = inputParams.Tags
	params.Requirements = inputParams.Requirements
	return nil
}

func (function Function) create(db *gorm.DB) error {
	result := db.Create(&function)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (function *Function) getAssociations(db *gorm.DB) {
	db.Model(&function).Association("Tags").Find(&function.Tags)
	db.Model(&function).Association("Requirements").Find(&function.Requirements)
}

func (function *Function) get(db *gorm.DB, inputFunction uint) error {
	result := db.Model(&Function{}).Where("id = ?", inputFunction).Take(&function)
	if result.Error != nil {
		return result.Error
	}
	function.getAssociations(db)
	return nil
}

func (function Function) update(db *gorm.DB) error {
	// Get the original Function object
	var originalFunction Function
	result := db.Model(&Function{}).Where("id = ?", function.Model.ID).Take(&originalFunction)
	if result.Error != nil {
		return result.Error
	}
	// Set the static values
	originalFunction.Name = function.Name
	// Set the associated values by grabbing them from the database
	tagsError := db.Model(&originalFunction).Association("Tags").Replace(&function.Tags)
	if tagsError != nil {
		return tagsError
	}
	requirementsError := db.Model(&originalFunction).Association("Requirements").Replace(&function.Requirements)
	if requirementsError != nil {
		return requirementsError
	}
	result = db.Model(&originalFunction).Update("name", originalFunction.Name)
	if result.Error != nil {
		err := errors.New("new name already exists")
		return err
	}
	db.Save(&originalFunction)
	return nil
}

func (function Function) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&function)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params FunctionParams) error {
	var err error
	var newTags []functiontag.FunctionTag
	if len(params.Tags) > 0 {
		newTags, err = functiontag.Get(db, params.Tags)
		if err != nil {
			return err
		}
	}
	var newRequirements []Function
	if len(params.Requirements) > 0 {
		newRequirements, err = Get(db, params.Requirements)
		if err != nil {
			return err
		}
	}
	return Function{
		Name:         params.Name,
		Tags:         newTags,
		Requirements: newRequirements,
	}.create(db)
}

func Get(db *gorm.DB, inputFunctions []uint) ([]Function, error) {
	var outputFunctions []Function
	if len(inputFunctions) < 1 {
		db.Model(&Function{}).Select("id").Find(&inputFunctions)
	}
	for _, inputFunction := range inputFunctions {
		var outputFunction Function
		err := outputFunction.get(db, inputFunction)
		if err != nil {
			return nil, err
		}
		outputFunctions = append(outputFunctions, outputFunction)
	}
	return outputFunctions, nil
}

func Update(db *gorm.DB, params FunctionParams) error {
	var err error
	var newTags []functiontag.FunctionTag
	if len(params.Tags) > 0 {
		newTags, err = functiontag.Get(db, params.Tags)
		if err != nil {
			return err
		}
	}
	var newRequirements []Function
	if len(params.Requirements) > 0 {
		newRequirements, err = Get(db, params.Requirements)
		if err != nil {
			return err
		}
	}
	return Function{
		Model:        gorm.Model{ID: params.IDArray[0]},
		Name:         params.Name,
		Tags:         newTags,
		Requirements: newRequirements,
	}.update(db)
}

func Delete(db *gorm.DB, inputFunctions []uint) error {
	var functions []Function
	// if len(inputFunctions) < 1 {
	// 	result := db.Model(&Function{}).Select("id").Find(&inputFunctions)
	// 	if result.Error != nil {
	// 		return result.Error
	// 	}
	// }
	for _, inputFunction := range inputFunctions {
		var function Function
		function.get(db, inputFunction)
		functions = append(functions, function)
	}
	for _, function := range functions {
		err := function.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) ([]Function, error) {
	var err error
	var params FunctionParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result []Function
	switch method {
	case "GET":
		result, err = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
