package functiontag

import (
	"encoding/json"
	"errors"

	"gorm.io/gorm"
)

type FunctionTag struct {
	gorm.Model
	Name string `gorm:"uniqueIndex" json:"name"`
}

type FunctionTagParams struct {
	// ID of the object being modified
	IDArray []uint
	// New fields
	Name string `json:"name"`
}

func (params *FunctionTagParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams FunctionTagParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	if inputParams.Name == "" {
		return errors.New("input name cannot be empty")
	}
	params.Name = inputParams.Name
	return nil
}

func (functionTag FunctionTag) create(db *gorm.DB) error {
	result := db.Create(&functionTag)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (functionTag *FunctionTag) get(db *gorm.DB, inputFunctionTag uint) error {
	result := db.Model(&FunctionTag{}).Where("ID = ?", inputFunctionTag).Take(&functionTag)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (functionTag FunctionTag) update(db *gorm.DB) error {
	var originalFunctionTag FunctionTag
	result := db.Model(&FunctionTag{}).Where("id = ?", functionTag.ID).Take(&originalFunctionTag)
	if result.Error != nil {
		return result.Error
	}
	originalFunctionTag.Name = functionTag.Name
	result = db.Model(&originalFunctionTag).Update("name", originalFunctionTag.Name)
	if result.Error != nil {
		err := errors.New("new name already exists")
		return err
	}
	db.Save(&originalFunctionTag)
	return nil
}

func (functionTag FunctionTag) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&functionTag)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params FunctionTagParams) error {
	return FunctionTag{
		Name: params.Name,
	}.create(db)
}

func Get(db *gorm.DB, inputFunctionTags []uint) ([]FunctionTag, error) {
	var outputFunctionTags []FunctionTag
	if len(inputFunctionTags) < 1 {
		db.Model(&FunctionTag{}).Select("id").Find(&inputFunctionTags)
	}
	for _, inputFunctionTag := range inputFunctionTags {
		var outputFunctionTag FunctionTag
		err := outputFunctionTag.get(db, inputFunctionTag)
		if err != nil {
			return nil, err
		}
		outputFunctionTags = append(outputFunctionTags, outputFunctionTag)
	}
	return outputFunctionTags, nil
}

func Update(db *gorm.DB, params FunctionTagParams) error {
	return FunctionTag{
		Model: gorm.Model{ID: params.IDArray[0]},
		Name:  params.Name,
	}.update(db)
}

func Delete(db *gorm.DB, inputFunctionTags []uint) error {
	var functionTags []FunctionTag
	// if len(inputFunctionTags) < 1 {
	// 	result := db.Model(&FunctionTag{}).Select("id").Find(&inputFunctionTags)
	// 	if result.Error != nil {
	// 		return result.Error
	// 	}
	// }
	for _, inputFunctionTag := range inputFunctionTags {
		var functionTag FunctionTag
		functionTag.get(db, inputFunctionTag)
		functionTags = append(functionTags, functionTag)
	}
	for _, functionTag := range functionTags {
		err := functionTag.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) ([]FunctionTag, error) {
	var err error
	var params FunctionTagParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result []FunctionTag
	switch method {
	case "GET":
		result, err = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
