package group

import (
	"encoding/json"
	"errors"

	"gorm.io/gorm"
)

type Group struct {
	gorm.Model
	Name string `gorm:"uniqueIndex" json:"name"`
}

type GroupParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	Name string `json:"name"`
}

func (params *GroupParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams GroupParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	if inputParams.Name == "" {
		return errors.New("input name cannot be empty")
	}
	params.Name = inputParams.Name
	return nil
}

func (group Group) create(db *gorm.DB) error {
	result := db.Create(&group)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (group *Group) get(db *gorm.DB, inputGroup uint) error {
	result := db.Model(&Group{}).Where("ID = ?", inputGroup).Take(&group)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (group Group) update(db *gorm.DB) error {
	var originalGroup Group
	result := db.Model(&Group{}).Where("id = ?", group.ID).Take(&originalGroup)
	if result.Error != nil {
		return result.Error
	}
	originalGroup.Name = group.Name
	result = db.Model(&originalGroup).Update("name", originalGroup.Name)
	if result.Error != nil {
		err := errors.New("new name already exists")
		return err
	}
	db.Save(&originalGroup)
	return nil
}

func (group Group) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&group)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params GroupParams) error {
	return Group{
		Name: params.Name,
	}.create(db)
}

func Get(db *gorm.DB, inputGroups []uint) ([]Group, error) {
	var outputGroups []Group
	if len(inputGroups) < 1 {
		db.Model(&Group{}).Select("id").Find(&inputGroups)
	}
	for _, inputGroup := range inputGroups {
		var outputGroup Group
		err := outputGroup.get(db, inputGroup)
		if err != nil {
			return nil, err
		}
		outputGroups = append(outputGroups, outputGroup)
	}
	return outputGroups, nil
}

func Update(db *gorm.DB, params GroupParams) error {
	return Group{
		Model: gorm.Model{ID: params.IDArray[0]},
		Name:  params.Name,
	}.update(db)
}

func Delete(db *gorm.DB, inputGroups []uint) error {
	var groups []Group
	// if len(inputGroups) < 1 {
	// 	result := db.Model(&Group{}).Select("id").Find(&inputGroups)
	// 	if result.Error != nil {
	// 		return result.Error
	// 	}
	// }
	for _, inputGroup := range inputGroups {
		var group Group
		err := group.get(db, inputGroup)
		if err != nil {
			return err
		}
		groups = append(groups, group)
	}
	for _, group := range groups {
		err := group.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) ([]Group, error) {
	var err error
	var params GroupParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result []Group
	switch method {
	case "GET":
		result, err = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
