package itemtag

import (
	"encoding/json"
	"errors"

	"gorm.io/gorm"
)

type ItemTag struct {
	gorm.Model
	Name string `gorm:"uniqueIndex" json:"name"`
}

type ItemTagParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	Name string `json:"name"`
}

func (params *ItemTagParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams ItemTagParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	if inputParams.Name == "" {
		return errors.New("input name cannot be empty")
	}
	params.Name = inputParams.Name
	return nil
}

func (itemTag ItemTag) create(db *gorm.DB) error {
	result := db.Create(&itemTag)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (itemTag *ItemTag) get(db *gorm.DB, inputItemTag uint) {
	db.Where("id = ?", inputItemTag).Take(&itemTag)
}

func (itemTag ItemTag) update(db *gorm.DB) error {
	// Get the original Function object
	var originalItemTag ItemTag
	result := db.Model(&ItemTag{}).Where("id = ?", itemTag.Model.ID).Take(&originalItemTag)
	if result.Error != nil {
		return result.Error
	}
	// Set the static values
	originalItemTag.Name = itemTag.Name
	result = db.Model(&originalItemTag).Update("name", originalItemTag.Name)
	if result.Error != nil {
		err := errors.New("new name already exists")
		return err
	}
	db.Save(&originalItemTag)
	return nil
}

func (itemTag ItemTag) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&itemTag)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params ItemTagParams) error {
	return ItemTag{
		Name: params.Name,
	}.create(db)
}

func Get(db *gorm.DB, inputItemTags []uint) *[]ItemTag {
	var outputItemTags []ItemTag
	if len(inputItemTags) < 1 {
		db.Model(&ItemTag{}).Select("id").Find(&inputItemTags)
	}
	for _, inputItemTag := range inputItemTags {
		var outputItemTag ItemTag
		outputItemTag.get(db, inputItemTag)
		outputItemTags = append(outputItemTags, outputItemTag)
	}
	return &outputItemTags
}

func Update(db *gorm.DB, params ItemTagParams) error {
	return ItemTag{
		Name: params.Name,
	}.update(db)
}

func Delete(db *gorm.DB, inputItemTags []uint) error {
	var itemTags []ItemTag
	// if len(inputItemTags) < 1 {
	// 	result := db.Model(&ItemTag{}).Select("id").Find(&inputItemTags)
	// 	if result.Error != nil {
	// 		return result.Error
	// 	}
	// }
	for _, inputItemTag := range inputItemTags {
		var itemTag ItemTag
		itemTag.get(db, inputItemTag)
		itemTags = append(itemTags, itemTag)
	}
	for _, itemTag := range itemTags {
		err := itemTag.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) (*[]ItemTag, error) {
	var err error
	var params ItemTagParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result *[]ItemTag
	switch method {
	case "GET":
		result = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
