package person

import (
	"encoding/json"
	"errors"

	group "example.com/database/group"

	"gorm.io/gorm"
)

type Person struct {
	gorm.Model
	Name   string        `gorm:"primaryKey uniqueIndex" json:"name"`
	Groups []group.Group `gorm:"many2many:person_group_associations" json:"groups"` // Unique
}

type PersonParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	Name   string `json:"name"`
	Groups []uint `json:"groups"`
}

func (params *PersonParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams PersonParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	if inputParams.Name == "" {
		return errors.New("input name cannot be empty")
	}
	params.Name = inputParams.Name
	params.Groups = inputParams.Groups
	return nil
}

func (person Person) create(db *gorm.DB) error {
	result := db.Create(&person)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (person *Person) getAssociations(db *gorm.DB) {
	db.Model(&person).Association("Groups").Find(&person.Groups)
}

func (person *Person) get(db *gorm.DB, inputPerson uint) {
	db.Where("id = ?", inputPerson).Take(&person)
	person.getAssociations(db)
}

func (person Person) update(db *gorm.DB) error {
	// Get the original Person object
	var originalPerson Person
	result := db.Model(&Person{}).Where("id = ?", person.Model.ID).Take(&originalPerson)
	if result.Error != nil {
		return result.Error
	}
	// Set the static values
	originalPerson.Name = person.Name
	// Set the associated values by grabbing them from the database
	groupsError := db.Model(&originalPerson).Association("Groups").Replace(&person.Groups)
	if groupsError != nil {
		return groupsError
	}
	result = db.Model(&originalPerson).Update("name", originalPerson.Name)
	if result.Error != nil {
		err := errors.New("new name already exists")
		return err
	}
	db.Save(&originalPerson)
	return nil
}

func (person Person) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&person)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params PersonParams) error {
	var newGroups []group.Group
	var err error
	if len(params.Groups) > 0 {
		newGroups, err = group.Get(db, params.Groups)
		if err != nil {
			return err
		}
	}
	return Person{
		Name:   params.Name,
		Groups: newGroups,
	}.create(db)
}

func GetByName(db *gorm.DB, name string) *Person {
	var person Person
	db.Model(&Person{}).Where("name = ?", name).Take(&person)
	person.getAssociations(db)
	return &person
}

func Get(db *gorm.DB, inputPersons []uint) *[]Person {
	var outputPersons []Person
	if len(inputPersons) < 1 {
		db.Model(&Person{}).Select("id").Find(&inputPersons)
	}
	for _, inputPerson := range inputPersons {
		var outputPerson Person
		outputPerson.get(db, inputPerson)
		outputPersons = append(outputPersons, outputPerson)
	}
	return &outputPersons
}

func Update(db *gorm.DB, params PersonParams) error {
	var newGroups []group.Group
	var err error
	if len(params.Groups) > 0 {
		newGroups, err = group.Get(db, params.Groups)
		if err != nil {
			return err
		}
	}
	return Person{
		Model:  gorm.Model{ID: params.IDArray[0]},
		Name:   params.Name,
		Groups: newGroups,
	}.update(db)
}

func Delete(db *gorm.DB, inputPersons []uint) error {
	var persons []Person
	if len(inputPersons) < 1 {
		result := db.Model(&Person{}).Select("id").Find(&inputPersons)
		if result.Error != nil {
			return result.Error
		}
	}
	for _, inputPerson := range inputPersons {
		var person Person
		person.get(db, inputPerson)
		persons = append(persons, person)
	}
	for _, person := range persons {
		err := person.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) (*[]Person, error) {
	var err error
	var params PersonParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result *[]Person
	switch method {
	case "GET":
		result = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
