package role

import (
	"encoding/json"
	"errors"

	group "example.com/database/group"
	tier "example.com/database/tier"

	"gorm.io/gorm"
)

type Role struct {
	gorm.Model
	Name       string        `gorm:"uniqueIndex" json:"name"`
	Tiers      []tier.Tier   `gorm:"many2many:role_tier_associations" json:"tiers"`
	Visibility []group.Group `gorm:"many2many:role_visibility_associations" json:"visibility"` // Unique
}

type RoleParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	Name       string `json:"name"`
	Tiers      []uint `json:"tiers"`
	Visibility []uint `json:"visibility"`
}

func (params *RoleParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams RoleParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	if inputParams.Name == "" {
		return errors.New("input name cannot be empty")
	}
	params.Name = inputParams.Name
	params.Tiers = inputParams.Tiers
	params.Visibility = inputParams.Visibility
	return nil
}

func (role Role) create(db *gorm.DB) error {
	result := db.Create(&role)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (role *Role) getAssociations(db *gorm.DB) {
	db.Model(&role).Association("Tiers").Find(&role.Tiers)
	db.Model(&role).Association("Visibility").Find(&role.Visibility)
}

func (role *Role) get(db *gorm.DB, inputRole uint) {
	db.Where("id = ?", inputRole).Take(&role)
	role.getAssociations(db)
}

func (role Role) update(db *gorm.DB) error {
	// Get the original Role object
	var originalRole Role
	result := db.Model(&Role{}).Where("id = ?", role.Model.ID).Take(&originalRole)
	if result.Error != nil {
		return result.Error
	}
	// Set the static values
	originalRole.Name = role.Name
	// Set the associated values by grabbing them from the database
	tiersError := db.Model(&originalRole).Association("Tiers").Replace(&role.Tiers)
	if tiersError != nil {
		return tiersError
	}
	visibilityError := db.Model(&originalRole).Association("Visibility").Replace(&role.Visibility)
	if visibilityError != nil {
		return visibilityError
	}
	result = db.Model(&originalRole).Update("name", originalRole.Name)
	if result.Error != nil {
		err := errors.New("new name already exists")
		return err
	}
	db.Save(&originalRole)
	return nil
}

func (role Role) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&role)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params RoleParams) error {
	var err error
	var newTiers []tier.Tier
	if len(params.Tiers) > 0 {
		newTiers = *tier.Get(db, params.Tiers)
	}
	var newVisibility []group.Group
	if len(params.Visibility) > 0 {
		newVisibility, err = group.Get(db, params.Visibility)
		if err != nil {
			return err
		}
	}
	return Role{
		Name:       params.Name,
		Tiers:      newTiers,
		Visibility: newVisibility,
	}.create(db)
}

func Get(db *gorm.DB, inputRoles []uint) *[]Role {
	var outputRoles []Role
	if len(inputRoles) < 1 {
		db.Model(&Role{}).Select("id").Find(&inputRoles)
	}
	for _, inputRole := range inputRoles {
		var outputRole Role
		outputRole.get(db, inputRole)
		outputRoles = append(outputRoles, outputRole)
	}
	return &outputRoles
}

func Update(db *gorm.DB, params RoleParams) error {
	var err error
	var newTiers []tier.Tier
	if len(params.Tiers) > 0 {
		newTiers = *tier.Get(db, params.Tiers)
	}
	var newVisibility []group.Group
	if len(params.Visibility) > 0 {
		newVisibility, err = group.Get(db, params.Visibility)
		if err != nil {
			return err
		}
	}
	return Role{
		Name:       params.Name,
		Tiers:      newTiers,
		Visibility: newVisibility,
	}.update(db)
}

func Delete(db *gorm.DB, inputRoles []uint) error {
	var roles []Role
	if len(inputRoles) < 1 {
		result := db.Model(&Role{}).Select("id").Find(&inputRoles)
		if result.Error != nil {
			return result.Error
		}
	}
	for _, inputRole := range inputRoles {
		var function Role
		function.get(db, inputRole)
		roles = append(roles, function)
	}
	for _, role := range roles {
		err := role.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) (*[]Role, error) {
	var err error
	var params RoleParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result *[]Role
	switch method {
	case "GET":
		result = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
