package schematic

import (
	"encoding/json"

	function "example.com/database/function"
	group "example.com/database/group"
	inventoryslot "example.com/database/inventoryslot"

	"gorm.io/gorm"
)

type Schematic struct {
	gorm.Model
	Material     []inventoryslot.InventorySlot `gorm:"many2many:schematic_material_associations" json:"material"`
	Tools        []inventoryslot.InventorySlot `gorm:"many2many:schematic_tool_associations" json:"tools"`
	Requirements []function.Function           `gorm:"many2many:schematic_requirement_associations" json:"requirements"`
	TimeUnits    int64                         `json:"time_units"` // Positive
	Result       inventoryslot.InventorySlot   `gorm:"many2many:schematic_result_associations" json:"result"`
	Visibility   []group.Group                 `gorm:"many2many:schematic_group_associations" json:"visibility"` // Unique
}

type SchematicParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	Material     []uint `json:"material"`
	Tools        []uint `json:"tools"`
	Requirements []uint `json:"requirements"`
	TimeUnits    int64  `json:"time_units"`
	Result       uint   `json:"result"`
	Visibility   []uint `json:"visibility"`
}

func (params *SchematicParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams SchematicParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	params.Material = inputParams.Material
	params.Tools = inputParams.Tools
	params.Requirements = inputParams.Requirements
	params.TimeUnits = inputParams.TimeUnits
	params.Result = inputParams.Result
	params.Visibility = inputParams.Visibility
	return nil
}

func (schematic Schematic) create(db *gorm.DB) error {
	result := db.Create(&schematic)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (schematic *Schematic) getAssociations(db *gorm.DB) {
	db.Model(&schematic).Association("Material").Find(&schematic.Material)
	db.Model(&schematic).Association("Tools").Find(&schematic.Tools)
	db.Model(&schematic).Association("Requirements").Find(&schematic.Requirements)
	db.Model(&schematic).Association("Result").Find(&schematic.Result)
	db.Model(&schematic).Association("Visibility").Find(&schematic.Visibility)
}

func (schematic *Schematic) get(db *gorm.DB, inputSchematic uint) {
	db.Model(&Schematic{}).Where("id = ?", inputSchematic).Take(&schematic)
	schematic.getAssociations(db)
}

func (schematic Schematic) update(db *gorm.DB) error {
	// Get the original Schematic object
	var originalSchematic Schematic
	result := db.Model(&Schematic{}).Where("id = ?", schematic.Model.ID).Take(&originalSchematic)
	if result.Error != nil {
		return result.Error
	}
	// Set the static values
	originalSchematic.TimeUnits = schematic.TimeUnits
	// Set the associated values by grabbing them from the database
	materialError := db.Model(&schematic).Association("Material").Replace(&schematic.Material)
	if materialError != nil {
		return materialError
	}
	toolsError := db.Model(&schematic).Association("Tools").Replace(&schematic.Tools)
	if toolsError != nil {
		return toolsError
	}
	requirementsError := db.Model(&schematic).Association("Requirements").Replace(&schematic.Requirements)
	if requirementsError != nil {
		return requirementsError
	}
	resultError := db.Model(&schematic).Association("Result").Replace(&schematic.Result)
	if resultError != nil {
		return resultError
	}
	visibilityError := db.Model(&schematic).Association("Visibility").Replace(&schematic.Visibility)
	if visibilityError != nil {
		return visibilityError
	}
	db.Save(&originalSchematic)
	return nil
}

func (schematic Schematic) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&schematic)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params SchematicParams) error {
	var err error
	var newMaterial []inventoryslot.InventorySlot
	if len(params.Material) > 0 {
		newMaterial = *inventoryslot.Get(db, params.Material)
	}
	var newTools []inventoryslot.InventorySlot
	if len(params.Tools) > 0 {
		newTools = *inventoryslot.Get(db, params.Tools)
	}
	var newRequirements []function.Function
	if len(params.Requirements) > 0 {
		newRequirements, err = function.Get(db, params.Requirements)
		if err != nil {
			return err
		}
	}
	resultArray := *inventoryslot.Get(db, []uint{params.Result})
	newResult := resultArray[0]
	var newVisibility []group.Group
	if len(params.Visibility) > 0 {
		newVisibility, err = group.Get(db, params.Visibility)
		if err != nil {
			return err
		}
	}
	return Schematic{
		Material:     newMaterial,
		Tools:        newTools,
		Requirements: newRequirements,
		TimeUnits:    params.TimeUnits,
		Result:       newResult,
		Visibility:   newVisibility,
	}.create(db)
}

func Get(db *gorm.DB, inputSchematics []uint) *[]Schematic {
	var outputSchematics []Schematic
	if len(inputSchematics) < 1 {
		db.Model(&Schematic{}).Select("id").Find(&inputSchematics)
	}
	for _, inputSchematic := range inputSchematics {
		var outputSchematic Schematic
		outputSchematic.get(db, inputSchematic)
		outputSchematics = append(outputSchematics, outputSchematic)
	}
	return &outputSchematics
}

func Update(db *gorm.DB, params SchematicParams) error {
	var err error
	var newMaterial []inventoryslot.InventorySlot
	if len(params.Material) > 0 {
		newMaterial = *inventoryslot.Get(db, params.Material)
	}
	var newTools []inventoryslot.InventorySlot
	if len(params.Tools) > 0 {
		newTools = *inventoryslot.Get(db, params.Tools)
	}
	var newRequirements []function.Function
	if len(params.Requirements) > 0 {
		newRequirements, err = function.Get(db, params.Requirements)
		if err != nil {
			return err
		}
	}
	resultArray := *inventoryslot.Get(db, []uint{params.Result})
	newResult := resultArray[0]
	var newVisibility []group.Group
	if len(params.Visibility) > 0 {
		newVisibility, err = group.Get(db, params.Visibility)
		if err != nil {
			return err
		}
	}
	return Schematic{
		Material:     newMaterial,
		Tools:        newTools,
		Requirements: newRequirements,
		TimeUnits:    params.TimeUnits,
		Result:       newResult,
		Visibility:   newVisibility,
	}.update(db)
}

func Delete(db *gorm.DB, inputSchematics []uint) error {
	var schematics []Schematic
	// if len(inputSchematics) < 1 {
	// 	result := db.Model(&Schematic{}).Select("id").Find(&inputSchematics)
	// 	if result.Error != nil {
	// 		return result.Error
	// 	}
	// }
	for _, inputSchematic := range inputSchematics {
		var schematic Schematic
		schematic.get(db, inputSchematic)
		schematics = append(schematics, schematic)
	}
	for _, schematic := range schematics {
		err := schematic.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) (*[]Schematic, error) {
	var err error
	var params SchematicParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result *[]Schematic
	switch method {
	case "GET":
		result = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
