package tier

import (
	"encoding/json"

	functionset "example.com/database/functionset"

	"gorm.io/gorm"
)

type Tier struct {
	gorm.Model
	FunctionSets []functionset.FunctionSet `gorm:"many2many:tier_functionset_associations" json:"function_sets"`
}

type TierParams struct {
	// ID(s) of the object being modified
	IDArray []uint
	// New fields
	FunctionSets []uint `json:"function_sets"`
}

func (params *TierParams) parse(IDUintArray *[]uint, body *[]byte) error {
	var inputParams TierParams
	params.IDArray = *IDUintArray
	if len(*body) == 0 {
		return nil
	}
	err := json.Unmarshal(*body, &inputParams)
	if err != nil {
		return err
	}
	params.FunctionSets = inputParams.FunctionSets
	return nil
}

func (tier Tier) create(db *gorm.DB) error {
	result := db.Create(&tier)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (tier *Tier) getAssociations(db *gorm.DB) {
	db.Model(&tier).Association("FunctionSets").Find(&tier.FunctionSets)
}

func (tier *Tier) get(db *gorm.DB, inputTier uint) {
	db.Where("id = ?", inputTier).Take(&tier)
	tier.getAssociations(db)
}

func (tier Tier) update(db *gorm.DB) error {
	// Get the original Tier object
	var originalTier Tier
	result := db.Model(&Tier{}).Where("id = ?", tier.Model.ID).Take(&originalTier)
	if result.Error != nil {
		return result.Error
	}
	// Set the associated values by grabbing them from the database
	tagsError := db.Model(&originalTier).Association("FunctionSets").Replace(&tier.FunctionSets)
	if tagsError != nil {
		return tagsError
	}
	db.Save(&originalTier)
	return nil
}

func (tier Tier) delete(db *gorm.DB) error {
	result := db.Unscoped().Delete(&tier)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, params TierParams) error {
	var newFunctionSets []functionset.FunctionSet
	var err error
	if len(params.FunctionSets) > 0 {
		newFunctionSets, err = functionset.Get(db, params.FunctionSets)
		if err != nil {
			return err
		}
	}
	return Tier{
		FunctionSets: newFunctionSets,
	}.create(db)
}

func Get(db *gorm.DB, inputTiers []uint) *[]Tier {
	var outputTiers []Tier
	if len(inputTiers) < 1 {
		db.Model(&Tier{}).Select("id").Find(&inputTiers)
	}
	for _, inputTier := range inputTiers {
		var outputTier Tier
		outputTier.get(db, inputTier)
		outputTiers = append(outputTiers, outputTier)
	}
	return &outputTiers
}

func Update(db *gorm.DB, params TierParams) error {
	var newFunctionSets []functionset.FunctionSet
	var err error
	if len(params.FunctionSets) > 0 {
		newFunctionSets, err = functionset.Get(db, params.FunctionSets)
		if err != nil {
			return err
		}
	}
	return Tier{
		Model:        gorm.Model{ID: params.IDArray[0]},
		FunctionSets: newFunctionSets,
	}.update(db)
}

func Delete(db *gorm.DB, inputTiers []uint) error {
	var tiers []Tier
	// if len(inputTiers) < 1 {
	// 	result := db.Model(&Tier{}).Select("id").Find(&inputTiers)
	// 	if result.Error != nil {
	// 		return result.Error
	// 	}
	// }
	for _, inputTier := range inputTiers {
		var tier Tier
		tier.get(db, inputTier)
		tiers = append(tiers, tier)
	}
	for _, tier := range tiers {
		err := tier.delete(db)
		if err != nil {
			return err
		}
	}
	return nil
}

func HandleRequest(method string, db *gorm.DB, IDUintArray *[]uint, body *[]byte) (*[]Tier, error) {
	var err error
	var params TierParams
	err = params.parse(IDUintArray, body)
	if err != nil {
		return nil, err
	}
	var result *[]Tier
	switch method {
	case "GET":
		result = Get(db, params.IDArray)
	case "POST":
		err = Create(db, params)
	case "PUT":
		err = Update(db, params)
	case "DELETE":
		err = Delete(db, params.IDArray)
	}
	return result, err
}
