package functiontagtest

import (
	"encoding/json"
	"fmt"
	"os"

	functiontag "example.com/database/functiontag"
	testrequest "example.com/test/testrequest"
	testsetup "example.com/test/testsetup"
	"github.com/gin-gonic/gin"
	"github.com/stretchr/testify/suite"
)

type TestFunctionTag struct {
	Result []functiontag.FunctionTag `json:"result"`
}

type FunctionTagTestSuite struct {
	suite.Suite
	globalAuthHeader  string
	router            *gin.Engine
	functiontag       functiontag.FunctionTag
	functiontagJSON   []byte
	functiontagUpdate functiontag.FunctionTag
}

func (s *FunctionTagTestSuite) SetupTestSuite() {
	s.globalAuthHeader, s.router = testsetup.SetupTestSuite()
	// Create Test Data
	fixtures, _ := os.ReadFile("fixtures/functiontags.json")
	var functionTagFixtures struct {
		FunctionTags      []functiontag.FunctionTag `json:"function_tags"`
		FunctionTagUpdate functiontag.FunctionTag   `json:"function_tag_update"`
	}
	json.Unmarshal(fixtures, &functionTagFixtures)
	s.functiontag = functionTagFixtures.FunctionTags[0]
	s.functiontagJSON, _ = json.Marshal(s.functiontag)
	s.functiontagUpdate = functionTagFixtures.FunctionTagUpdate
}

func (s *FunctionTagTestSuite) TearDownTestSuite() {
	os.Remove("db/main.db")
}

func (s *FunctionTagTestSuite) Test01CreateFunctionTag() {
	// Setup variables
	var output TestFunctionTag
	target := fmt.Sprintf("/function-tag?id=%d", s.functiontag.ID)

	// Attempt to create functiontag
	status := testrequest.MakePostRequest(&s.functiontagJSON, "/function-tag", s.globalAuthHeader, s.router)
	// Check that the request was successful
	s.Equal(status, "200 OK", "function tag created successfully")

	// Attempt to get functiontag
	body := testrequest.MakeGetRequest(target, s.router)
	// Unmarshal the result
	json.Unmarshal(body, &output)
	// Check that the functiontag listed matches the created functiontag
	s.Equal(s.functiontag.Name, output.Result[0].Name, "input name matches output name")
}

func (s *FunctionTagTestSuite) Test02GetFunctionTag() {
	// Setup variables
	var output TestFunctionTag

	// Attempt to get functiontags
	body := testrequest.MakeGetRequest("/function-tag", s.router)
	// Unmarshal the result
	json.Unmarshal(body, &output)
	// Check that the functiontags listed match the created functiontags
	s.Equal(s.functiontag.Name, output.Result[0].Name, "input name matches output name")
}

func (s *FunctionTagTestSuite) Test03UpdateFunctionTag() {
	// Setup variables
	var output TestFunctionTag
	body, _ := json.Marshal(s.functiontagUpdate)
	target := fmt.Sprintf("/function-tag?id=%d", s.functiontag.ID)

	// Attempt to update functiontag
	status := testrequest.MakePutRequest(&body, target, s.globalAuthHeader, s.router)
	// Check that the request was successful
	s.Equal(status, "200 OK", "function tag created successfully")

	// Attempt to get functiontag
	body = testrequest.MakeGetRequest(target, s.router)
	// Unmarshal the result
	json.Unmarshal(body, &output)
	// Check that the functiontag listed matches the updated functiontag
	s.Equal(s.functiontagUpdate.Name, output.Result[0].Name, "input name matches output name")
}

func (s *FunctionTagTestSuite) Test04DeleteFunctionTag() {
	// Setup variables
	var output TestFunctionTag
	target := fmt.Sprintf("/function-tag?id=%d", s.functiontag.ID)

	// Attempt to delete functiontag
	status := testrequest.MakeDeleteRequest(target, s.globalAuthHeader, s.router)
	// Check that the request was successful
	s.Equal(status, "200 OK", "function tag deleted successfully")

	// Attempt to get functiontag
	body := testrequest.MakeGetRequest(target, s.router)
	// Unmarshal the result
	json.Unmarshal(body, &output)
	// Check that the functiontag list is empty
	s.Equal(len(output.Result), 0, "input name matches output name")
}
